@@ -44,7 +44,7 @@
 #include "msettings.h"
 #include "partman.h"
 
-#define PARTMAN_SAFETY_MB 32 // 1MB at start + Compensate for rounding errors.
+#define PARTMAN_SAFETY_MB 8 // 1MB at start + Compensate for rounding errors.
 #define PARTMAN_MAX_PARTS 128 // Maximum number of partitions Linux supports.
 
 PartMan::PartMan(MProcess &mproc, Ui::MeInstall &ui, QWidget *parent)
@@ -98,11 +98,20 @@ void PartMan::scan(DeviceItem *drvstart)
     const QJsonArray &jsonBD = jsonObjBD["blockdevices"].toArray();
 
     if (!drvstart) {
-        proc.exec("partprobe -s", true);
-        proc.exec("blkid -c /dev/null", true);
-        root.clear();
+        proc.exec("partprobe -s", true); // Inform the kernel of partition changes.
+        proc.exec("blkid -c /dev/null", true); // Refresh cached blkid information.
+    }
+    // Partitions listed in order of their physical locations.
+    QStringList order;
+    QString curdev;
+    for(const QString &line : proc.execOutLines("parted -lm", true)) {
+        const QString &sect = line.section(':', 0, 0);
+        const int part = sect.toInt();
+        if (part) order.append(DeviceItem::join(curdev, part));
+        else if (sect.startsWith("/dev/")) curdev = sect.mid(5);
     }
 
+    if (!drvstart) root.clear();
     for (const QJsonValue &jsonDrive : jsonBD) {
         const QString &jdName = jsonDrive["name"].toString();
         if (jsonDrive["type"] != "disk") continue;
@@ -112,6 +121,7 @@ void PartMan::scan(DeviceItem *drvstart)
         else if (jdName != drvstart->device) continue;
 
         drvit->clear();
+        drvit->flags.curEmpty = true;
         drvit->flags.oldLayout = true;
         drvit->device = jdName;
         drvit->flags.useGPT = (jsonDrive["pttype"]=="gpt");
@@ -122,10 +132,13 @@ void PartMan::scan(DeviceItem *drvstart)
         const QJsonArray &jsonParts = jsonDrive["children"].toArray();
         for (const QJsonValue &jsonPart : jsonParts) {
             const QString &partTypeName = jsonPart["parttypename"].toString();
-            if (partTypeName=="Extended") continue;
+            if (partTypeName == "Extended" || partTypeName == "W95 Ext'd (LBA)"
+                || partTypeName == "Linux extended") continue;
             DeviceItem *partit = new DeviceItem(DeviceItem::Partition, drvit);
+            drvit->flags.curEmpty = false;
             partit->flags.oldLayout = true;
             partit->device = jsonPart["name"].toString();
+            partit->order = order.indexOf(partit->device);
             partit->size = jsonPart["size"].toVariant().toLongLong();
             partit->physec = jsonPart["phy-sec"].toInt();
             partit->curLabel = jsonPart["label"].toString();
@@ -133,20 +146,23 @@ void PartMan::scan(DeviceItem *drvstart)
             const int partflags = jsonPart["partflags"].toString().toUInt(nullptr, 0);
             if ((partflags & 0x80) || (partflags & 0x04)) partit->setActive(true);
             partit->mapCount = jsonPart["children"].toArray().count();
-            partit->flags.curESP = (partTypeName=="EFI System");
-            if (partit->flags.curESP) partit->usefor = "ESP";
-            if (!partit->flags.nasty) partit->flags.nasty = partTypeName.startsWith("Microsoft LDM");
+            partit->flags.curESP = partTypeName.startsWith("EFI "); // "System"/"(FAT-12/16/32)"
+            if (partTypeName.startsWith("Microsoft LDM")) partit->flags.nasty = true;
             partit->flags.bootRoot = (!bootUUID.isEmpty() && jsonPart["uuid"]==bootUUID);
             partit->curFormat = jsonPart["fstype"].toString();
             // Propagate the boot and nasty flags up to the drive.
             if (partit->flags.bootRoot) drvit->flags.bootRoot = true;
             if (partit->flags.nasty) drvit->flags.nasty = true;
-            notifyChange(partit);
+            else if (partit->flags.curESP && partit->allowedUsesFor().contains("ESP")) {
+                partit->usefor = "ESP";
+                partit->format = partit->allowedFormats().at(0);
+            }
         }
         for (int ixPart = drvit->childCount() - 1; ixPart >= 0; --ixPart) {
             // Propagate the boot flag across the entire drive.
             if (drvit->flags.bootRoot) drvit->child(ixPart)->flags.bootRoot = true;
         }
+        drvit->sortChildren();
         notifyChange(drvit);
     }
 
@@ -215,16 +231,19 @@ void PartMan::scanVirtualDevices(bool rescan)
             devit->curLabel = label;
             devit->curFormat = jsonDev["fstype"].toString();
             devit->flags.cryptoV = crypto;
-            notifyChange(devit);
+            devit->flags.oldLayout = true;
         }
+    }
+    for (const auto &it : listed.toStdMap()) delete it.second;
+    vdlit->sortChildren();
+    for(int ixi = 0; ixi < vdlit->childCount(); ++ixi) {
+        const QString &name = vdlit->child(ixi)->device;
         DeviceItem *orit = findOrigin(name);
         if (orit) {
             orit->devMapper = name;
             notifyChange(orit);
         }
     }
-    for (const auto &it : listed.toStdMap()) delete it.second;
-    vdlit->sortChildren();
     gui.treePartitions->expand(index(vdlit));
 }
 
@@ -237,19 +256,13 @@ bool PartMan::manageConfig(MSettings &config, bool save)
         int partCount = drvit->childCount();
         bool drvPreserve = drvit->flags.oldLayout;
         const QString &configNewLayout = "Storage/NewLayout." + drvit->device;
-        QVariant bootdev;
         if (save) {
             if (!drvPreserve) config.setValue(configNewLayout, partCount);
-            if (drvit->active) {
-                bootdev = drvit->active->device;
-                config.setValue("Boot", bootdev);
-            }
         } else if (config.contains(configNewLayout)) {
             drvPreserve = false;
             drvit->clear();
             partCount = config.value(configNewLayout).toInt();
             if (partCount > PARTMAN_MAX_PARTS) return false;
-            bootdev = config.value("Boot");
         }
         // Partition configuration.
         const long long sizeMax = drvit->size - (PARTMAN_SAFETY_MB * 1048576);
@@ -258,7 +271,7 @@ bool PartMan::manageConfig(MSettings &config, bool save)
             DeviceItem *partit = nullptr;
             if (save || drvPreserve) {
                 partit = drvit->child(ixPart);
-                partit->flags.oldLayout = drvPreserve;
+                if (!save) partit->flags.oldLayout = true;
             } else {
                 partit = new DeviceItem(DeviceItem::Partition, drvit);
                 partit->device = DeviceItem::join(drvit->device, ixPart+1);
@@ -284,16 +297,15 @@ bool PartMan::manageConfig(MSettings &config, bool save)
                     if (sizeTotal > sizeMax) return false;
                     if (config.value("Boot").toBool()) partit->setActive(true);
                 }
-                partit->usefor = config.value("UseFor").toString();
-                partit->format = config.value("Format").toString();
-                partit->chkbadblk = config.value("CheckBadBlocks").toBool();
-                partit->encrypt = config.value("Encrypt").toBool();
-                partit->label = config.value("Label").toString();
-                partit->options = config.value("Options").toString();
-                partit->dump = config.value("Dump").toBool();
-                partit->pass = config.value("Pass").toInt();
+                partit->usefor = config.value("UseFor", partit->usefor).toString();
+                partit->format = config.value("Format", partit->format).toString();
+                partit->chkbadblk = config.value("CheckBadBlocks", partit->chkbadblk).toBool();
+                partit->encrypt = config.value("Encrypt", partit->encrypt).toBool();
+                partit->label = config.value("Label", partit->label).toString();
+                partit->options = config.value("Options", partit->options).toString();
+                partit->dump = config.value("Dump", partit->dump).toBool();
+                partit->pass = config.value("Pass", partit->pass).toInt();
             }
-            if (bootdev == partit->device) partit->setActive(true);
             int subvolCount = 0;
             if (partit->format == "btrfs") {
                 if (!save) subvolCount = config.value("Subvolumes").toInt();
@@ -365,7 +377,7 @@ void PartMan::treeSelChange()
         bool islocked = true;
         if (isdrive) {
             islocked = twit->isLocked();
-            if (twit->childCount() <= 0) isold = false;
+            if (twit->flags.curEmpty) isold = false;
         }
         gui.pushPartClear->setEnabled(!islocked);
         gui.pushPartRemove->setEnabled(!isold && !isdrive);
@@ -460,7 +472,7 @@ void PartMan::treeMenu(const QPoint &)
 
         const bool locked = twit->isLocked();
         actClear->setDisabled(locked);
-        actReset->setDisabled(locked || twit->flags.oldLayout);
+        actReset->setDisabled(locked);
         menuTemplates->setDisabled(locked);
         actCrypto->setVisible(gui.boxCryptoPass->isVisible());
 
@@ -470,8 +482,9 @@ void PartMan::treeMenu(const QPoint &)
         else if (action == actBasic) twit->layoutDefault(-1, false);
         else if (action == actCrypto) twit->layoutDefault(-1, true);
         else if (action == actReset) {
-            while (twit->childCount()) delete twit->child(0);
+            gui.boxMain->setEnabled(false);
             scan(twit);
+            gui.boxMain->setEnabled(true);
         }
     } else if (twit->type == DeviceItem::Subvolume) {
         QAction *actRemSubvolume = menu.addAction(tr("Remove subvolume"));
@@ -704,7 +717,7 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
         for (int ixdrv = 0; ixdrv < root.childCount(); ++ixdrv) {
             DeviceItem *drvit = root.child(ixdrv);
             const int partCount = drvit->childCount();
-            bool setupGPT = gptoverride || drvit->size >= 2199023255552 || partCount > 4;
+            bool setupGPT = uefi || gptoverride || drvit->size >= 2199023255552 || partCount > 4;
             if (drvit->flags.oldLayout) {
                 setupGPT = drvit->flags.useGPT;
             } else if (drvit->type != DeviceItem::VirtualDevices) {
@@ -751,30 +764,6 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
         if (msgbox.exec() != QMessageBox::Yes) return false;
     }
 
-    calculatePartBD();
-    return true;
-}
-
-bool PartMan::calculatePartBD()
-{
-    listToUnmount.clear();
-    for (int ixDrive = 0; ixDrive < root.childCount(); ++ixDrive) {
-        DeviceItem *drvit = root.child(ixDrive);
-        if (drvit->type == DeviceItem::VirtualDevices) continue;
-        const bool useExist = drvit->flags.oldLayout;
-        const int partCount = drvit->childCount();
-        const long long driveSize = drvit->size / 1048576;
-        if (!useExist) {
-            // see if GPT needs to be used (either UEFI or >=2TB drive)
-            drvit->flags.useGPT = gptoverride || uefi || driveSize >= 2097152 || partCount > 4;
-        }
-        // Add (or mark) future partitions to the list.
-        for (int ixPart=0; ixPart < partCount; ++ixPart) {
-            DeviceItem *twit = drvit->child(ixPart);
-            if (twit->usefor.isEmpty()) continue;
-            if (useExist) listToUnmount << twit->device;
-        }
-    }
     return true;
 }
 
@@ -900,7 +889,25 @@ bool PartMan::preparePartitions()
 {
     proc.log(__PRETTY_FUNCTION__);
 
-    // detach all existing partitions on the selected drive
+    // Detach all existing partitions.
+    QStringList listToUnmount;
+    for (int ixDrive = 0; ixDrive < root.childCount(); ++ixDrive) {
+        DeviceItem *drvit = root.child(ixDrive);
+        if (drvit->type == DeviceItem::VirtualDevices) continue;
+        const int partCount = drvit->childCount();
+        if (drvit->flags.oldLayout) {
+            // Using the existing layout, so only mark used partitions for unmounting.
+            for (int ixPart=0; ixPart < partCount; ++ixPart) {
+                DeviceItem *twit = drvit->child(ixPart);
+                if (!twit->usefor.isEmpty()) listToUnmount << twit->device;
+            }
+        } else {
+            // Clearing the drive, so mark all partitions on the drive for unmounting.
+            listToUnmount << proc.execOutLines("lsblk -nro name /dev/" + drvit->device, true);
+            // See if GPT needs to be used (either UEFI or >=2TB drive)
+            drvit->flags.useGPT = (gptoverride || uefi || drvit->size >= 2199023255552 || partCount > 4);
+        }
+    }
     for (const QString &strdev : listToUnmount) {
         proc.exec("swapoff /dev/" + strdev, true);
         proc.exec("/bin/umount /dev/" + strdev, true);
@@ -976,7 +983,6 @@ bool PartMan::preparePartitions()
         }
     }
     proc.exec("partprobe -s", true);
-    proc.sleep(1000);
     return true;
 }
 
@@ -1002,7 +1008,10 @@ bool PartMan::formatPartitions()
         if (useFor == "FORMAT") proc.status(fmtstatus.arg(dev));
         else proc.status(fmtstatus.arg(twit->shownUseFor()));
         if (useFor == "ESP") {
-            if (!proc.exec("mkfs.msdos -F " + twit->format.mid(3)+' '+dev)) return false;
+            QString cmd("mkfs.msdos -F " + twit->format.mid(3));
+            if (twit->chkbadblk) cmd.append(" -c");
+            if (!twit->label.isEmpty()) cmd += " -n \"" + twit->label.trimmed().left(11) + '\"';
+            if (!proc.exec(cmd + ' ' + dev)) return false;
             // Sets boot flag and ESP flag.
             const QStringList &devsplit = DeviceItem::split(dev);
             if (!proc.exec("parted -s /dev/" + devsplit.at(0)
@@ -1444,10 +1453,9 @@ Qt::ItemFlags PartMan::flags(const QModelIndex &index) const
         if (item->allowedUsesFor().count() > 1) flagsOut |= Qt::ItemIsEditable;
         break;
     case Label:
-        if (item->type == DeviceItem::Subvolume) {
-            if (item->format != "PRESERVE") flagsOut |= Qt::ItemIsEditable;
-        } else {
-            if (!item->usefor.isEmpty()) flagsOut |= Qt::ItemIsEditable;
+        if (item->format != "PRESERVE") {
+            if (item->type == DeviceItem::Subvolume) flagsOut |= Qt::ItemIsEditable;
+            else if (!item->usefor.isEmpty()) flagsOut |= Qt::ItemIsEditable;
         }
         break;
     case Encrypt:
@@ -1457,7 +1465,8 @@ Qt::ItemFlags PartMan::flags(const QModelIndex &index) const
         if (item->allowedFormats().count() > 1) flagsOut |= Qt::ItemIsEditable;
         break;
     case Check:
-        if (item->format.startsWith("ext") || item->format == "jfs") {
+        if (item->format.startsWith("ext") || item->format == "jfs"
+            || item->format.startsWith("FAT", Qt::CaseInsensitive)) {
             flagsOut |= Qt::ItemIsUserCheckable;
         }
         break;
@@ -1590,10 +1599,9 @@ int PartMan::changeEnd(bool notify)
 }
 void PartMan::notifyChange(class DeviceItem *item, int first, int last)
 {
-    DeviceItem *const p = item->parentItem;
     if (first < 0) first = 0;
-    if (last < 0) last = p ? p->childCount() - 1 : 0;
-    const int row = p ? p->indexOfChild(item) : 0;
+    if (last < 0) last = _TreeColumns_ - 1;
+    const int row = item->row();
     emit dataChanged(createIndex(row, first, item), createIndex(row, last, item));
 }
 
@@ -1604,6 +1612,7 @@ DeviceItem::DeviceItem(enum DeviceType type, DeviceItem *parent, DeviceItem *pre
 {
     if (type == Partition) size = 1048576;
     if (parent) {
+        physec = parent->physec;
         partman = parent->partman;
         const int i = preceding ? (parentItem->children.indexOf(preceding) + 1) : parentItem->childCount();
         if (partman) partman->beginInsertRows(partman->index(parent), i, i);
@@ -1671,6 +1680,7 @@ inline int DeviceItem::childCount() const
 void DeviceItem::sortChildren()
 {
     auto cmp = [](DeviceItem *l, DeviceItem *r) {
+        if (l->order != r->order) return l->order < r->order;
         return l->device < r->device;
     };
     std::sort(children.begin(), children.end(), cmp);
@@ -1762,7 +1772,7 @@ QStringList DeviceItem::allowedUsesFor(bool real) const
         list.prepend("Format");
         if (type != VirtualBD) {
             if (size <= 16777216) list << "BIOS-GRUB";
-            if (size <= 4294967296) list << "ESP" << "boot";
+            if (size <= 8589934592) list << "ESP" << "boot";
         }
         list << "swap" << "home";
     }
@@ -1776,18 +1786,20 @@ QStringList DeviceItem::allowedFormats() const
     if (!isVolume()) return QStringList();
     QStringList list;
     bool allowPreserve = false, selPreserve = false;
-    if (type == Subvolume) list << "CREATE";
+    if (type == Subvolume) list.append("CREATE");
     else {
         const QString &use = realUseFor();
         if (use.isEmpty()) return QStringList();
-        else if (use == "/boot") list << "ext4";
-        else if (use == "BIOS-GRUB") list << "GRUB";
+        else if (use == "/boot") list.append("ext4");
+        else if (use == "BIOS-GRUB") list.append("GRUB");
         else if (use == "ESP") {
-            list << "FAT32" << "FAT16" << "FAT12";
+            list.append("FAT32");
+            if (size <= 4294901760) list.append("FAT16");
+            if (size <= 33553920) list.append("FAT12");
             selPreserve = allowPreserve = (list.contains(curFormat, Qt::CaseInsensitive)
                 || !curFormat.compare("VFAT", Qt::CaseInsensitive));
         } else if (use == "SWAP") {
-            list << "SWAP";
+            list.append("SWAP");
             selPreserve = allowPreserve = list.contains(curFormat, Qt::CaseInsensitive);
         } else {
             list << "ext4" << "ext3" << "ext2";
@@ -2096,7 +2108,16 @@ QWidget *DeviceItemDelegate::createEditor(QWidget *parent, const QStyleOptionVie
     QComboBox *combo = nullptr;
     switch (index.column())
     {
-    case PartMan::Size: widget = new QSpinBox(parent); break;
+    case PartMan::Size:
+        {
+            QSpinBox *spin = new QSpinBox(parent);
+            spin->setSuffix("MB");
+            spin->setStepType(QSpinBox::AdaptiveDecimalStepType);
+            spin->setAccelerated(true);
+            spin->setSpecialValueText("MAX");
+            widget = spin;
+        }
+        break;
     case PartMan::UseFor:
         widget = combo = new QComboBox(parent);
         combo->setEditable(true);
@@ -2134,17 +2155,14 @@ void DeviceItemDelegate::setEditorData(QWidget *editor, const QModelIndex &index
     case PartMan::Size:
         {
             QSpinBox *spin = qobject_cast<QSpinBox *>(editor);
-            spin->setRange(1, static_cast<int>(item->driveFreeSpace() / 1048576));
-            spin->setSuffix("MB");
-            spin->setStepType(QSpinBox::AdaptiveDecimalStepType);
-            spin->setAccelerated(true);
+            spin->setRange(0, static_cast<int>(item->driveFreeSpace() / 1048576));
             spin->setValue(item->size / 1048576);
-            spin->setWrapping(true);
         }
         break;
     case PartMan::UseFor:
         {
             QComboBox *combo = qobject_cast<QComboBox *>(editor);
+            combo->clear();
             combo->addItem("");
             combo->addItems(item->allowedUsesFor(false));
             combo->setCurrentText(item->usefor);
@@ -2156,6 +2174,7 @@ void DeviceItemDelegate::setEditorData(QWidget *editor, const QModelIndex &index
     case PartMan::Format:
         {
             QComboBox *combo = qobject_cast<QComboBox *>(editor);
+            combo->clear();
             const QStringList &formats = item->allowedFormats();
             assert(!formats.isEmpty());
             for (const QString &fmt : formats) {
@@ -2188,6 +2207,7 @@ void DeviceItemDelegate::setModelData(QWidget *editor, QAbstractItemModel *model
     case PartMan::Size:
         item->size = qobject_cast<QSpinBox *>(editor)->value();
         item->size *= 1048576; // Separate step to prevent int overflow.
+        if (item->size == 0) item->size = item->driveFreeSpace();
         break;
     case PartMan::UseFor:
         item->usefor = qobject_cast<QComboBox *>(editor)->currentText().trimmed();
